/* HashFile.c
 * Michael McTernan, mm7323@bris.ac.uk
 * 23/3/00
 * An ADT for disk storage.
 * Completely generic so can be reused.
 * Uses double hashing, with second hash func being
 *  limited to 2<f'(s)<40 so that there won't
 *  be huge seeks when doing a probe sequence.
 *  Also small seeks avoided as it would have to
 *  wait for a disk spin.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "HashFile.h"

/* I usually have the following as a function to allow breakpoint debugging */
#define Error(a)  { fprintf(stderr,"%s %d: %s\n",__FILE__,__LINE__,a);\
                    perror(NULL); \
                    abort(); }

/* The ADT structure */
struct hashfile {
  FILE *f;
  long  n;
  int   RecordLength;
};

/* The disk record structure */
typedef struct hashrecord {
  char   Free;
  short  KeyLength;
} hashrecord_t;

/* Prototypes */
static int isPrime(int v);
static int KeyEqual(const void *Key1,int KeyLength1,const void *Key2,int KeyLength2);
static int HashFunction1(struct hashfile *HF,const void *Key,int KeyLength);
static int HashFunction2(struct hashfile *HF,const void *Key,int KeyLength);


/*HASH_New*/
/* Create a new hashfile
 */
struct hashfile *HASH_New(char *File, long Records, int RecordLength) {
  struct hashfile *NewHF;
  hashrecord_t    *Record;
  int              n;
  
  assert(isPrime(Records));

  /* Allocate some memory */
  if((NewHF=malloc(sizeof(struct hashfile)))==NULL)
    Error("malloc failed");

  /* Open the file and fill in the fields */
  if((NewHF->f=fopen(File,"w+b"))==NULL)
    Error("fopen failed");

  NewHF->n=Records;
  NewHF->RecordLength=RecordLength+sizeof(hashrecord_t);

  /* Now create a blank record */
  if((Record=calloc(NewHF->RecordLength,1))==NULL)
    Error("calloc failed");
    
  Record->Free=1;
  
  /* Now create the blank records */
  for(n=0;n<Records;n++) {
    if(fwrite(Record,NewHF->RecordLength,1,NewHF->f)!=1)
      Error("fwrite failed");
  }

  free(Record);
  return NewHF;
}


/*HASH_Open*/
/* Open a hash file
 */
struct hashfile *HASH_Open(char *File,int ReadOnly,int RecordLength) {
  struct hashfile *NewHF;

  /* Allocate some memory */
  if((NewHF=malloc(sizeof(struct hashfile)))==NULL)
    Error("malloc failed");

  /* Open the file */
  if(ReadOnly) {
    if((NewHF->f=fopen(File,"rb"))==NULL)
      Error("fopen failed");
  } else {
    if((NewHF->f=fopen(File,"r+b"))==NULL)
      Error("fopen failed");
  }

  /* Calculate how many elements are in the file */
  fseek(NewHF->f,0,SEEK_END);
  NewHF->n=ftell(NewHF->f)/(RecordLength+sizeof(hashrecord_t));
  NewHF->RecordLength=RecordLength+sizeof(hashrecord_t);

  if(!isPrime(NewHF->n))
    Error("Invalid hashtable: Size is not a prime number");

  /* Return */
  return NewHF;
}


/*HASH_Get*/
/* Retrive somthing from the table
 */
int HASH_Get(struct hashfile *HF,const void *Key,int KeyLength,void *Result) {
  hashrecord_t *Record;
  long          RecordNum,s;

  /* Make some space for the record */
  if((Record=malloc(HF->RecordLength))==NULL)
    Error("malloc failed");
    
  /* Find where it should be */
  s=RecordNum=HashFunction1(HF,Key,KeyLength);

  fseek(HF->f,RecordNum*HF->RecordLength,SEEK_SET);
  
  if(fread(Record,HF->RecordLength,1,HF->f)!=1)
    Error("fread failed");

  while(Record->Free==0 &&
        !KeyEqual(Key,KeyLength,&Record[1],Record->KeyLength))
  {

    /* Check the chain */
    RecordNum=(RecordNum+HashFunction2(HF,Key,KeyLength)) % HF->n;
    fseek(HF->f,RecordNum*HF->RecordLength,SEEK_SET);
  
    if(fread(Record,HF->RecordLength,1,HF->f)!=1)
      Error("fread failed");

    /* Check if we have searched the entire table */
    if(RecordNum==s) {
      free(Record);
      return 0;
    }
    
  }

  /* Check if it was found */
  if(Record->Free) {
    free(Record);
    return 0;
  }
  
  /* We have it! */
  memcpy(Result,&Record[1],HF->RecordLength-sizeof(hashrecord_t));

  free(Record);
  
  return 1;

}


/*HASH_Add*/
/* Add a record to the hash table
 */
void HASH_Add(struct hashfile *HF,const void *Data,int KeyLength) {
  hashrecord_t   *DiskRecord,*Record;
  long            RecordNum,m;

  /* Allocate memory for the records */
  DiskRecord=malloc(HF->RecordLength*100);
  Record=malloc(HF->RecordLength*100);
  
  if(DiskRecord==NULL || Record==NULL)
    Error("malloc failed");

  Record->Free=0;
  Record->KeyLength=KeyLength;
  
  memcpy(&Record[1],Data,
         HF->RecordLength-sizeof(hashrecord_t));
    
  /* Check for a free slot */
  m=RecordNum=HashFunction1(HF,&Record[1],KeyLength);

  fseek(HF->f,HF->RecordLength*RecordNum,SEEK_SET);
  if(fread(DiskRecord,HF->RecordLength,1,HF->f)!=1)
      Error("fread failed");

  /* Search for a free slot */
  while(!DiskRecord->Free) {
      
    RecordNum=(RecordNum+HashFunction2(HF,&Record[1],strlen((char *)&Record[1]))) % HF->n;
    fseek(HF->f,HF->RecordLength*RecordNum,SEEK_SET);
      
    if(fread(DiskRecord,HF->RecordLength,1,HF->f)!=1)
        Error("fread failed");

    /* Check if we have looped the whole table */
    if(RecordNum==m)
      Error("Hash table full");
  } 

  fseek(HF->f,HF->RecordLength*RecordNum,SEEK_SET);

  if(fwrite(Record,HF->RecordLength,1,HF->f)!=1)
    Error("fwrite failed");

  free(Record);
  free(DiskRecord);

}

/*HASH_Close*/
/* Close the previously opened hashfile
 */
void HASH_Close(struct hashfile *HF) {
  fclose(HF->f);
  free(HF);
}

/*HASH_CalculateNiceSize*/
/* Given the maximum number of records and
 *  an ideal loading, return the smallest
 *  prime number that will give the ideal
 *  loading.
 */
long HASH_CalculateNiceSize(long MaxRecords,float IdealLoadFactor) {

  /* l=n/m === m=n/l */
  MaxRecords/=IdealLoadFactor;

  while(!isPrime(MaxRecords))
    MaxRecords++;

  return MaxRecords;
}


/*HashFunction1*/
static int HashFunction1(struct hashfile *HF,const void *Key,int KeyLength) {
  char *String=(char *)Key;
  int Total=0,t=0;

  for(t=0;t<KeyLength;t++) {
    Total+=String[t]*7;
  }

  return Total % HF->n;
  
}


/*HashFunction2*/
static int HashFunction2(struct hashfile *HF,const void *Key,int KeyLength) {
  char *String=(char *)Key;
  int Total=0,t=KeyLength;

  for(t=0;t<KeyLength;t++) {
    Total+=String[t]*3;
  }

  Total*=7;

  return (Total % 40)+3;
  
}


/*isPrime*/
static int isPrime(int v) {
  int w;

  for(w=2;w<v;w++)
    if((float)((float)v/(float)w)==(float)((int)v/(int)w))
      return 0;

  return 1;

}

/*KeyEqual*/
/* Like memcmp, except checks the length of the keys
 *  to be the same first
 */
static int KeyEqual(const void *Key1,int KeyLength1,const void *Key2,int KeyLength2) {

  if(KeyLength1!=KeyLength2)
    return 0;

  return memcmp(Key1,Key2,KeyLength1)==0;

}
